"""File generated while packaging."""
import contextlib

version_info = {'short': '0.2.1', 'full': '0.2.1', 'release': True, 'clean': True, 'git revision': 'a4b6d5f80589840eadf8a5a9b9908565ba06bb57'}
version = '0.2.1'


@contextlib.contextmanager
def version_manager():
    """Dummy context manager, returns the version."""
    yield version
